"""
Convert a RIB file to a list of /24 prefixes.
TODO: IPv6 support.
"""
from ipaddress import IPv4Network, ip_network
from pathlib import Path

import typer
from pyasn.mrtx import parse_mrt_file
from tqdm import tqdm


def main(
    input_file: Path = typer.Argument(...),
    output_file: Path = typer.Argument(...),
    print_progress: bool = False,
    skip_record_on_error: bool = False,
) -> None:
    prefixes = parse_mrt_file(str(input_file), print_progress, skip_record_on_error)
    with output_file.open("w") as f:
        for prefix in tqdm(prefixes, disable=not print_progress):
            net = ip_network(prefix)
            if isinstance(net, IPv4Network):
                if net.prefixlen <= 24:
                    for subnet in net.subnets(new_prefix=24):
                        f.write(f"{subnet}\n")


def run() -> None:
    """
    Run the CLI from a path like `zeph_utils.zeph_bgp_convert:run`.
    Useful for poetry.
    """
    typer.run(main)


if __name__ == "__main__":
    run()



from abc import ABC, abstractmethod

from zeph.typing import Agent, Link, Network


class AbstractRanker(ABC):
    @abstractmethod
    def __call__(
        self, links: dict[tuple[Agent, Network], set[Link]]
    ) -> dict[Agent, list[Network]]:
        ...



from zeph.rankers.abstract import AbstractRanker
from zeph.rankers.dfg import DFGCoverRanker
from zeph.rankers.greedy import GreedyCoverRanker
from zeph.rankers.naive import NaiveRanker
from zeph.rankers.unique import UniqueLinksRanker

__all__ = (
    "AbstractRanker",
    "DFGCoverRanker",
    "GreedyCoverRanker",
    "NaiveRanker",
    "UniqueLinksRanker",
)



"""API drivers."""
from io import BytesIO
from uuid import uuid4

from httpx import HTTPStatusError
from iris_client import IrisClient

from zeph.typing import Network


def create_measurement(client: IrisClient, definition: dict) -> dict:
    res = client.post("/measurements", json=definition)
    try:
        res.raise_for_status()
    except HTTPStatusError as e:
        raise RuntimeError(res.content) from e
    return dict(res.json())  # make mypy happy...


def get_agents(client: IrisClient, tag: str) -> dict[str, dict]:
    agents = client.all("/agents", params=dict(tag=tag))
    return {agent["uuid"]: agent for agent in agents}


def get_measurement_agents(client: IrisClient, measurement_uuid: str) -> list[str]:
    measurement = client.get(f"/measurements/{measurement_uuid}").json()
    return [agent["agent_uuid"] for agent in measurement["agents"]]


def upload_prefix_list(
    client: IrisClient,
    prefixes: set[Network],
    protocol: str,
    min_ttl: int,
    max_ttl: int,
) -> str:
    key = f"zeph__{uuid4()}.csv"
    file = BytesIO(
        "\n".join(
            f"{prefix},{protocol},{min_ttl},{max_ttl},6" for prefix in prefixes
        ).encode()
    )
    res = client.post("/targets", files={"target_file": (key, file)})
    try:
        res.raise_for_status()
    except HTTPStatusError as e:
        raise RuntimeError(res.content) from e
    return key



def measurement_id(measurement_uuid: str, agent_uuid: str) -> str:
    """
    Return the measurement identifier used by Iris.
    >>> measurement_id("measurement-1", "agent-1")
    'measurement_1__agent_1'
    """
    return f"{measurement_uuid.replace('-', '_')}__{agent_uuid.replace('-', '_')}"


def parse_network(addr: str, prefix_len_v4: int = 24, prefix_len_v6: int = 64) -> str:
    """
    Convert an IPv6 address (as stored in the database) to a network string.
    If `addr` is an IPv4-mapped IPv6 address (starting with `::ffff:`) this
    will return an IPv4 network; otherwise it will return an IPv6 network.
    >>> parse_network("2001:db8::")
    '2001:db8::/64'
    >>> parse_network("::ffff:192.0.2.0")
    '192.0.2.0/24'
    >>> parse_network("::ffff:c000:200")
    Traceback (most recent call last):
    AssertionError: IPv4-mapped IPv6 addresses must be in dotted representation
    """
    if addr.startswith("::ffff:"):
        assert (
            "." in addr
        ), "IPv4-mapped IPv6 addresses must be in dotted representation"
        return f"{addr[7:]}/{prefix_len_v4}"
    return f"{addr}/{prefix_len_v6}"



from dataclasses import dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.query import LinksQuery, links_table
from diamond_miner.typing import IPNetwork
from pych_client import ClickHouseClient

from zeph.typing import Agent, Link, Network
from zeph.utilities import measurement_id, parse_network


@dataclass(frozen=True)
class GetUniqueLinksByPrefix(LinksQuery):
    """
    Get the list of unique links per prefix.
    To make things faster, we only retrieve the hash of each link,
    since we do not need the actual IP addresses seen.
    """

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"""
        SELECT
            probe_dst_prefix,
            groupUniqArray(cityHash64((near_addr, far_addr))) AS links
        FROM {links_table(measurement_id)}
        WHERE {self.filters(subset)}
        GROUP BY probe_dst_prefix
        """

    def for_all_agents(
        self, client: ClickHouseClient, measurement_uuid: str, agents_uuid: list[str]
    ) -> dict[tuple[Agent, Network], set[Link]]:
        links: dict[tuple[Agent, Network], set[Link]] = {}
        for agent_uuid in agents_uuid:
            for row in self.execute_iter(
                client, measurement_id(measurement_uuid, agent_uuid)
            ):
                network = parse_network(row["probe_dst_prefix"])
                links[(agent_uuid, network)] = set(row["links"])
        return links



from collections import defaultdict
from math import floor, log

from zeph.rankers import AbstractRanker
from zeph.typing import Agent, Link, Network


class DFGCoverRanker(AbstractRanker):
    """
    Reference:
    Cormode, Graham, Howard Karloff, and Anthony Wirth.
    "Set cover algorithms for very large datasets."
    Proceedings of the 19th ACM international conference on Information and knowledge management. 2010.
    """

    def __init__(self, p: float = 1.05):
        self.p = p

    def __call__(
        self, links: dict[tuple[Agent, Network], set[Link]]
    ) -> dict[Agent, list[Network]]:
        if not links:
            return {}

        covered: set[Link] = set()
        prefixes = defaultdict(list)
        subcollections = defaultdict(list)

        # Populate the sub-collections
        for (agent, prefix), links_ in links.items():
            k = floor(log(len(links_) or 1, self.p))
            subcollections[k].append((agent, prefix))
        k_max = max(subcollections.keys())

        # k = k_max ... 1
        for k in range(k_max, 0, -1):
            for (agent, prefix) in subcollections[k]:
                if len(links[(agent, prefix)] - covered) >= self.p**k:
                    prefixes[agent].append(prefix)
                    covered.update(links[(agent, prefix)])
                else:
                    links[(agent, prefix)] -= covered
                    k_prime = floor(log(len(links[(agent, prefix)]) or 1, self.p))
                    subcollections[k_prime].append((agent, prefix))

        # k = 0
        for (agent, prefix) in subcollections[0]:
            if len(links[(agent, prefix)] - covered) == 1:
                prefixes[agent].append(prefix)
                covered.update(links[(agent, prefix)])

        return prefixes



"""Base for selectors."""

import random
from abc import ABC, abstractmethod

from zeph.typing import Network


class AbstractSelector(ABC):
    def __init__(self, universe: set[Network], budgets: dict[str, int]) -> None:
        self.universe = universe
        self.budgets = budgets

    def universe_shuffled(self) -> list[Network]:
        universe = list(self.universe)
        random.shuffle(universe)
        return universe

    @abstractmethod
    def select(self, agent_uuid: str) -> set[Network]:
        pass

    def _select_random(
        self, agent_uuid: str, preset: set[Network] | None = None
    ) -> set[Network]:
        prefixes = set()
        if preset:
            prefixes.update(preset)
        universe = self.universe_shuffled()
        budget = self.budgets[agent_uuid]
        for prefix in universe:
            if len(prefixes) >= budget:
                break
            prefixes.add(prefix)
        return prefixes



from zeph.selectors.abstract import AbstractSelector
from zeph.selectors.constrained import ConstrainedRandomSelector
from zeph.selectors.epsilon import EpsilonSelector
from zeph.selectors.random import RandomSelector

__all__ = (
    "AbstractSelector",
    "ConstrainedRandomSelector",
    "EpsilonSelector",
    "RandomSelector",
)
